import re
import kubernetes
from hardeneks import helpers
from hardeneks.rules import Rule, Result
from hardeneks import Resources


class check_EKS_version(Rule):
    _type = "cluster_wide"
    pillar = "scalability"
    section = "control_plane"
    message = "EKS Version Should be greater or equal to 1.24."
    url = "https://aws.github.io/aws-eks-best-practices/scalability/docs/control-plane/#use-eks-124-or-above"

    def check(self, resources: Resources):
        client = kubernetes.client.VersionApi()
        version = client.get_code()
        minor = version.minor

        if int(re.sub("[^0-9]", "", minor)) < 24:
            self.result = Result(
                status=False,
                resources=f"{version.major}.{minor}",
                resource_type="Cluster Version",
            )
        else:
            self.result = Result(status=True, resource_type="Cluster Version")


#
# check_kubectl_compression
# checks all clusters in config for disable-compression flag set to true
# if any cluster does not have setting, it returns False
class check_kubectl_compression(Rule):
    _type = "cluster_wide"
    pillar = "scalability"
    section = "control_plane"
    message = "`disable-compression` in kubeconfig should equal True"
    url = "https://aws.github.io/aws-eks-best-practices/scalability/docs/control-plane/#disable-kubectl-compression"

    def check(self, resources: Resources):
        kubeconfig = helpers.get_kube_config()
        for cluster in kubeconfig.get("clusters", []):
            clusterName = cluster.get("name", "")
            if resources.cluster in clusterName:
                if not (
                    cluster.get("cluster", {}).get(
                        "disable-compression", False
                    )
                ):
                    self.result = Result(
                        status=False, resource_type="Compression Setting"
                    )
                else:
                    self.result = Result(
                        status=True, resource_type="Compression Setting"
                    )
                break



import os
from pathlib import Path
from pkg_resources import resource_filename
import tempfile
import yaml
import json
from collections import defaultdict

from botocore.exceptions import EndpointConnectionError
import boto3
import kubernetes
from rich.console import Console
from rich.table import Table
from rich.panel import Panel
import typer

from .resources import (
    NamespacedResources,
    Resources,
)
from .harden import harden
from hardeneks import helpers

app = typer.Typer()
console = Console(record=True)


def _config_callback(value: str):

    config = Path(value)

    if config.is_dir():
        raise typer.BadParameter(f"{config} is a directory")
    elif not config.exists():
        raise typer.BadParameter(f"{config} doesn't exist")

    with open(value, "r") as f:
        try:
            yaml.safe_load(f)
        except yaml.YAMLError as exc:
            raise typer.BadParameter(exc)

    return value


def _get_current_context(context):
    if context:
        return context
    _, active_context = kubernetes.config.list_kube_config_contexts()
    return active_context["name"]


def _get_namespaces(ignored_ns: list) -> list:
    v1 = kubernetes.client.CoreV1Api()
    namespaces = [i.metadata.name for i in v1.list_namespace().items]
    return list(set(namespaces) - set(ignored_ns))


def _get_cluster_name(context, region):
    try:
        client = boto3.client("eks", region_name=region)
        for name in client.list_clusters()["clusters"]:
            if name in context:
                return name
    except EndpointConnectionError:
        raise ValueError(f"{region} seems like a bad region name")


def _get_region():
    return boto3.session.Session().region_name


def _add_tls_verify():
    kubeconfig = helpers.get_kube_config()
    tmp_config = tempfile.NamedTemporaryFile().name

    for cluster in kubeconfig["clusters"]:
        cluster["cluster"]["insecure-skip-tls-verify"] = True
    with open(tmp_config, "w") as fd:
        yaml.dump(kubeconfig, fd, default_flow_style=False)

    kubernetes.config.load_kube_config(tmp_config)
    os.remove(tmp_config)


def _export_json(rules: list, json_path=str):
    def ndd():
        return defaultdict(ndd)

    json_blob = ndd()

    for rule in rules:
        result = {
            "status": rule.result.status,
            "resources": rule.result.resources,
            "resource_type": rule.result.resource_type,
            "namespace": rule.result.namespace,
            "resolution": rule.url,
        }
        json_blob[rule._type][rule.pillar][rule.section][rule.message] = result
    with open(json_path, "w", encoding="utf-8") as f:
        json.dump(json_blob, f, ensure_ascii=False, indent=4)


def print_consolidated_results(rules: list):

    pillars = set([i.pillar for i in rules])

    for pillar in pillars:
        table = Table()
        table.add_column("Section")
        table.add_column("Namespace")
        table.add_column("Rule")
        table.add_column("Resource")
        table.add_column("Resource Type")
        table.add_column("Resolution")
        filtered_rules = [i for i in rules if i.pillar == pillar]
        for rule in filtered_rules:
            color = "red"
            namespace = "Cluster Wide"
            if rule.result.status:
                color = "green"
            if rule.result.namespace:
                namespace = rule.result.namespace
            for resource in rule.result.resources:
                table.add_row(
                    rule.section,
                    namespace,
                    rule.message,
                    resource,
                    rule.result.resource_type,
                    f"[link={rule.url}]Link[/link]",
                    style=color,
                )
        console.print(Panel(table, title=f"[cyan][bold]{pillar} rules"))
        console.print()


@app.command()
def run_hardeneks(
    region: str = typer.Option(
        default=None, help="AWS region of the cluster. Ex: us-east-1"
    ),
    context: str = typer.Option(
        default=None,
        help="K8s context.",
    ),
    cluster: str = typer.Option(default=None, help="Cluster name."),
    namespace: str = typer.Option(
        default=None,
        help="Specific namespace to harden. Default is all namespaces.",
    ),
    config: str = typer.Option(
        default=resource_filename(__name__, "config.yaml"),
        callback=_config_callback,
        help="Path to a hardeneks config file.",
    ),
    export_txt: str = typer.Option(
        default=None,
        help="Export the report in txt format",
    ),
    export_html: str = typer.Option(
        default=None,
        help="Export the report in html format",
    ),
    export_json: str = typer.Option(
        default=None, help="Export the report in json format"
    ),
    insecure_skip_tls_verify: bool = typer.Option(
        False,
        "--insecure-skip-tls-verify",
    ),
    width: int = typer.Option(
        default=None, help="Width of the console (defaults to terminal width)"
    ),
    height: int = typer.Option(
        default=None,
        help="Height of the console (defaults to terminal height)",
    ),
):
    """
    Main entry point to hardeneks.

    Args:
        region (str): AWS region of the cluster. Ex: us-east-1
        context (str): K8s context
        cluster (str): Cluster name
        namespace (str): Specific namespace to be checked
        config (str): Path to hardeneks config file
        export-txt (str): Export the report in txt format
        export-html (str): Export the report in html format
        export-json (str): Export the report in json format
        insecure-skip-tls-verify (str): Skip tls verification
        width (int): Output width
        height (int): Output height

    Returns:
        None

    """
    if insecure_skip_tls_verify:
        _add_tls_verify()
    else:
        # should pass in config file
        kubernetes.config.load_kube_config(context=context)

    if width:
        console.width = width
    if height:
        console.height = height

    context = _get_current_context(context)

    if not cluster:
        cluster = _get_cluster_name(context, region)

    if not region:
        region = _get_region()

    console.rule("[b]HARDENEKS", characters="*  ")
    console.print(f"You are operating at {region}")
    console.print(f"You context is {context}")
    console.print(f"Your cluster name is {cluster}")
    console.print(f"You are using {config} as your config file")
    console.print()

    with open(config, "r") as f:
        config = yaml.safe_load(f)

    if not namespace:
        namespaces = _get_namespaces(config["ignore-namespaces"])
    else:
        namespaces = [namespace]

    rules = config["rules"]

    resources = Resources(region, context, cluster, namespaces)
    resources.set_resources()

    results = []

    cluster_wide_results = harden(resources, rules, "cluster_wide")

    results = results + cluster_wide_results

    for ns in namespaces:
        resources = NamespacedResources(region, context, cluster, ns)
        resources.set_resources()
        namespace_based_results = harden(resources, rules, "namespace_based")
        results = results + namespace_based_results

    print_consolidated_results(results)

    if export_txt:
        console.save_text(export_txt)
    if export_html:
        console.save_html(export_html)
    if export_json:
        _export_json(results, export_json)



from kubernetes import client


class Resources:
    def __init__(self, region, context, cluster, namespaces):
        self.region = region
        self.context = context
        self.cluster = cluster
        self.namespaces = namespaces

    def set_resources(self):
        self.cluster_roles = (
            client.RbacAuthorizationV1Api().list_cluster_role().items
        )
        self.cluster_role_bindings = (
            client.RbacAuthorizationV1Api().list_cluster_role_binding().items
        )
        self.resource_quotas = (
            client.CoreV1Api().list_resource_quota_for_all_namespaces().items
        )
        self.network_policies = (
            client.NetworkingV1Api()
            .list_network_policy_for_all_namespaces()
            .items
        )
        self.storage_classes = client.StorageV1Api().list_storage_class().items
        self.persistent_volumes = (
            client.CoreV1Api().list_persistent_volume().items
        )


class NamespacedResources:
    def __init__(self, region, context, cluster, namespace):
        self.namespace = namespace
        self.region = region
        self.cluster = cluster
        self.context = context

    def set_resources(self):
        self.roles = (
            client.RbacAuthorizationV1Api()
            .list_namespaced_role(self.namespace)
            .items
        )
        self.pods = (
            client.CoreV1Api().list_namespaced_pod(self.namespace).items
        )
        self.role_bindings = (
            client.RbacAuthorizationV1Api()
            .list_namespaced_role_binding(self.namespace)
            .items
        )
        self.deployments = (
            client.AppsV1Api().list_namespaced_deployment(self.namespace).items
        )
        self.daemon_sets = (
            client.AppsV1Api().list_namespaced_daemon_set(self.namespace).items
        )
        self.stateful_sets = (
            client.AppsV1Api()
            .list_namespaced_stateful_set(self.namespace)
            .items
        )
        self.services = (
            client.CoreV1Api().list_namespaced_service(self.namespace).items
        )
        self.hpas = (
            client.AutoscalingV1Api()
            .list_namespaced_horizontal_pod_autoscaler(self.namespace)
            .items
        )



import boto3

from ...resources import Resources
from hardeneks.rules import Rule, Result


class check_logs_are_enabled(Rule):
    _type = "cluster_wide"
    pillar = "security"
    section = "detective_controls"
    message = "Enable control plane logs for auditing."
    url = "https://aws.github.io/aws-eks-best-practices/security/docs/detective/#enable-audit-logs"

    def check(self, resources: Resources):
        client = boto3.client("eks", region_name=resources.region)
        cluster_metadata = client.describe_cluster(name=resources.cluster)
        logs = cluster_metadata["cluster"]["logging"]["clusterLogging"][0][
            "enabled"
        ]
        self.result = Result(status=True, resource_type="Log Configuration")
        if not logs:
            self.result = Result(
                status=False, resource_type="Log Configuration"
            )



from hardeneks.rules import Rule, Result
from hardeneks.resources import NamespacedResources


class use_encryption_with_aws_load_balancers(Rule):
    _type = "namespace_based"
    pillar = "security"
    section = "network_security"
    message = "Make sure you specify an ssl cert."
    url = "https://aws.github.io/aws-eks-best-practices/security/docs/network/#use-encryption-with-aws-load-balancers"

    def check(self, namespaced_resources: NamespacedResources):
        offenders = []
        for service in namespaced_resources.services:
            annotations = service.metadata.annotations
            if annotations:
                ssl_cert = (
                    "service.beta.kubernetes.io/aws-load-balancer-ssl-cert"
                    in annotations
                )
                ssl_cert_port = annotations.get(
                    "service.beta.kubernetes.io/aws-load-balancer-ssl-ports"
                )
                if not (ssl_cert and ssl_cert_port == "443"):
                    offenders.append(service)

        self.result = Result(status=True, resource_type="Service")
        if offenders:
            self.result = Result(
                status=False,
                resource_type="Service",
                resources=[i.metadata.name for i in offenders],
                namespace=namespaced_resources.namespace,
            )



from ...resources import NamespacedResources
from hardeneks.rules import Rule, Result


class disallow_linux_capabilities(Rule):
    _type = "namespace_based"
    pillar = "security"
    section = "runtime_security"
    message = "Capabilities beyond the allowed list are disallowed."
    url = "https://aws.github.io/aws-eks-best-practices/security/docs/runtime/#consider-adddropping-linux-capabilities-before-writing-seccomp-policies"

    def check(self, namespaced_resources: NamespacedResources):
        offenders = []

        allowed_list = [
            "AUDIT_WRITE",
            "CHOWN",
            "DAC_OVERRIDE",
            "FOWNER",
            "FSETID",
            "KILL",
            "MKNOD",
            "NET_BIND_SERVICE",
            "SETFCAP",
            "SETGID",
            "SETPCAP",
            "SETUID",
            "SYS_CHROOT",
        ]
        for pod in namespaced_resources.pods:
            for container in pod.spec.containers:
                if (
                    container.security_context
                    and container.security_context.capabilities
                ):
                    capabilities = set(
                        container.security_context.capabilities.add
                    )
                    if not capabilities.issubset(set(allowed_list)):
                        offenders.append(pod)

        self.result = Result(status=True, resource_type="Pod")
        if offenders:
            self.result = Result(
                status=False,
                resource_type="Pod",
                resources=[i.metadata.name for i in offenders],
                namespace=namespaced_resources.namespace,
            )



import boto3

from ...resources import Resources
from hardeneks.rules import Rule, Result


class use_immutable_tags_with_ecr(Rule):
    _type = "cluster_wide"
    pillar = "security"
    section = "image_security"
    message = "Make image tags immutable."
    url = "https://aws.github.io/aws-eks-best-practices/security/docs/image/#use-immutable-tags-with-ecr"

    def check(self, resources: Resources):
        offenders = []

        client = boto3.client("ecr", region_name=resources.region)
        repositories = client.describe_repositories()
        for repository in repositories["repositories"]:
            if repository["imageTagMutability"] != "IMMUTABLE":
                offenders.append(repository)

        self.result = Result(status=True, resource_type="ECR Repository")
        if offenders:
            self.result = Result(
                status=False,
                resource_type="ECR Repository",
                resources=[i["repositoryName"] for i in offenders],
            )



from kubernetes import client

from hardeneks.rules import Rule, Result
from hardeneks.resources import Resources


class check_metrics_server_is_running(Rule):
    _type = "cluster_wide"
    pillar = "reliability"
    section = "applications"
    message = "Metrics server is not deployed."
    url = "https://aws.github.io/aws-eks-best-practices/reliability/docs/application/#run-kubernetes-metrics-server"

    def check(self, resources: Resources):
        services = [
            i.metadata.name
            for i in client.CoreV1Api().list_service_for_all_namespaces().items
        ]

        if "metrics-server" in services:
            self.result = Result(status=True, resource_type="Service")
        else:
            self.result = Result(status=False, resource_type="Service")


class check_vertical_pod_autoscaler_exists(Rule):
    _type = "cluster_wide"
    pillar = "reliability"
    section = "applications"
    message = "Vertical pod autoscaler is not deployed."
    url = "https://aws.github.io/aws-eks-best-practices/reliability/docs/application/#run-kubernetes-metrics-server"

    def check(self, resources: Resources):

        deployments = [
            i.metadata.name
            for i in client.AppsV1Api()
            .list_deployment_for_all_namespaces()
            .items
        ]

        if "vpa-recommender" in deployments:
            self.result = Result(status=True, resource_type="Deployment")
        else:
            self.result = Result(status=False, resource_type="Deployment")



import boto3

from ...resources import Resources
from hardeneks.rules import Rule, Result


class deploy_workers_onto_private_subnets(Result):
    _type = "cluster_wide"
    pillar = "security"
    section = "infrastructure_security"
    message = "Place worker nodes on private subnets."
    url = "https://aws.github.io/aws-eks-best-practices/security/docs/hosts/#deploy-workers-onto-private-subnets"

    def check(self, resources: Resources):
        client = boto3.client("ec2", region_name=resources.region)

        offenders = []

        instance_metadata = client.describe_instances(
            Filters=[
                {
                    "Name": "tag:aws:eks:cluster-name",
                    "Values": [
                        resources.cluster,
                    ],
                },
            ]
        )

        for instance in instance_metadata["Reservations"]:
            if instance["Instances"][0]["PublicDnsName"]:
                offenders.append(instance["Instances"][0]["InstanceId"])

        self.result = Result(status=True, resource_type="Node")

        if offenders:
            self.result = Result(
                status=False, resource_type="Node", resources=offenders
            )


class make_sure_inspector_is_enabled(Rule):
    _type = "cluster_wide"
    pillar = "security"
    section = "infrastructure_security"
    message = "Enable Amazon Inspector for ec2 and ecr."
    url = "https://aws.github.io/aws-eks-best-practices/security/docs/hosts/#deploy-workers-onto-private-subnets"

    def check(self, resources: Resources):
        client = boto3.client("inspector2", region_name=resources.region)
        account_id = boto3.client(
            "sts", region_name=resources.region
        ).get_caller_identity()["Account"]

        response = client.batch_get_account_status(
            accountIds=[
                account_id,
            ]
        )

        resource_state = response["accounts"][0]["resourceState"]
        ec2_status = resource_state["ec2"]["status"]
        ecr_status = resource_state["ecr"]["status"]

        self.result = Result(
            status=True, resource_type="Inspector Configuration"
        )

        if ec2_status != "ENABLED" and ecr_status != "ENABLED":
            self.result = Result(
                status=False, resource_type="Inspector Configuration"
            )



from importlib import import_module


def harden(resources, config, _type):
    config = config[_type]
    results = []
    for pillar in config.keys():
        for section in config[pillar]:
            for rule in config[pillar][section]:
                module = import_module(f"hardeneks.{_type}.{pillar}.{section}")
                try:
                    cls = getattr(module, rule)
                except AttributeError as exc:
                    print(f"[bold][red]{exc}")
                try:
                    rule = cls()
                    rule.check(resources)
                    results.append(rule)
                except Exception as exc:
                    print(f"[bold][red]{exc}")

    return results



import boto3
from kubernetes import client

from ...resources import Resources
from hardeneks.rules import Rule, Result


class check_vpc_flow_logs(Rule):
    _type = "cluster_wide"
    pillar = "security"
    section = "network_security"
    message = "Enable flow logs for your VPC."
    url = "https://aws.github.io/aws-eks-best-practices/security/docs/network/#log-network-traffic-metadata"

    def check(self, resources: Resources):
        client = boto3.client("eks", region_name=resources.region)
        cluster_metadata = client.describe_cluster(name=resources.cluster)

        vpc_id = cluster_metadata["cluster"]["resourcesVpcConfig"]["vpcId"]
        client = boto3.client("ec2", region_name=resources.region)

        flow_logs = client.describe_flow_logs(
            Filters=[{"Name": "resource-id", "Values": [vpc_id]}]
        )["FlowLogs"]

        self.result = Result(status=True, resource_type="VPC Configuration")
        if not flow_logs:
            self.result = Result(
                status=False, resource_type="VPC Configuration"
            )


class check_awspca_exists(Rule):
    _type = "cluster_wide"
    pillar = "security"
    section = "network_security"
    message = "Install aws privateca issuer for your certificates."
    url = "https://aws.github.io/aws-eks-best-practices/security/docs/network/#acm-private-ca-with-cert-manager"

    def check(self, resources: Resources):
        services = client.CoreV1Api().list_service_for_all_namespaces().items
        for service in services:
            if service.metadata.name.startswith("aws-privateca-issuer"):
                self.result = Result(status=True, resource_type="Service")

        self.result = Result(
            status=False,
            resource_type="Service",
            resources=["aws-privateca-issuer"],
        )


class check_default_deny_policy_exists(Rule):
    _type = "cluster_wide"
    pillar = "security"
    section = "network_security"
    message = "Namespaces that does not have default network deny policies."
    url = "https://aws.github.io/aws-eks-best-practices/security/docs/network/#create-a-default-deny-policy"

    def check(self, resources: Resources):
        offenders = resources.namespaces

        for policy in resources.network_policies:
            offenders.remove(policy.metadata.namespace)

        self.result = Result(status=True, resource_type="Namespace")

        if offenders:
            self.result = Result(
                status=False, resource_type="Service", resources=offenders
            )



from ...resources import NamespacedResources
from hardeneks.rules import Rule, Result


class avoid_running_singleton_pods(Rule):
    _type = "namespace_based"
    pillar = "reliability"
    section = "applications"
    message = "Avoid running pods without deployments."
    url = "https://aws.github.io/aws-eks-best-practices/reliability/docs/application/#avoid-running-singleton-pods"

    def check(self, namespaced_resources: NamespacedResources):
        offenders = []
        for pod in namespaced_resources.pods:
            owner = pod.metadata.owner_references
            if not owner:
                offenders.append(pod)

        self.result = Result(status=True, resource_type="Pod")
        if offenders:
            self.result = Result(
                status=False,
                resource_type="Pod",
                resources=[i.metadata.name for i in offenders],
                namespace=namespaced_resources.namespace,
            )


class run_multiple_replicas(Rule):
    _type = "namespace_based"
    pillar = "reliability"
    section = "applications"
    message = "Avoid running single replica deployments."
    url = "https://aws.github.io/aws-eks-best-practices/reliability/docs/application/#run-multiple-replicas"

    def check(self, namespaced_resources: NamespacedResources):

        offenders = []

        for deployment in namespaced_resources.deployments:
            if deployment.spec.replicas < 2:
                offenders.append(deployment)

        self.result = Result(status=True, resource_type="Deployment")
        if offenders:
            self.result = Result(
                status=False,
                resource_type="Deployment",
                resources=[i.metadata.name for i in offenders],
                namespace=namespaced_resources.namespace,
            )


class schedule_replicas_across_nodes(Rule):
    _type = "namespace_based"
    pillar = "reliability"
    section = "applications"
    message = "Spread replicas across AZs and Nodes."
    url = "https://aws.github.io/aws-eks-best-practices/reliability/docs/application/#schedule-replicas-across-nodes"

    def check(self, namespaced_resources: NamespacedResources):

        offenders = []

        for deployment in namespaced_resources.deployments:
            spread = deployment.spec.template.spec.topology_spread_constraints
            if not spread:
                offenders.append(deployment)
            else:
                topology_keys = set([i.topology_key for i in spread])
                if not set(["topology.kubernetes.io/zone"]).issubset(
                    topology_keys
                ):
                    offenders.append(deployment)

        self.result = Result(status=True, resource_type="Deployment")
        if offenders:
            self.result = Result(
                status=False,
                resource_type="Deployment",
                resources=[i.metadata.name for i in offenders],
                namespace=namespaced_resources.namespace,
            )


class check_horizontal_pod_autoscaling_exists(Rule):
    _type = "namespace_based"
    pillar = "reliability"
    section = "applications"
    message = "Deploy horizontal pod autoscaler for deployments."
    url = "https://aws.github.io/aws-eks-best-practices/reliability/docs/application/#horizontal-pod-autoscaler-hpa"

    def check(self, namespaced_resources: NamespacedResources):

        offenders = []

        hpas = [
            i.spec.scale_target_ref.name for i in namespaced_resources.hpas
        ]

        for deployment in namespaced_resources.deployments:
            if deployment.metadata.name not in hpas:
                offenders.append(deployment)

        self.result = Result(status=True, resource_type="Deployment")
        if offenders:
            self.result = Result(
                status=False,
                resource_type="Deployment",
                resources=[i.metadata.name for i in offenders],
                namespace=namespaced_resources.namespace,
            )


class check_readiness_probes(Rule):
    _type = "namespace_based"
    pillar = "reliability"
    section = "applications"
    message = "Define readiness probes for pods."
    url = "https://aws.github.io/aws-eks-best-practices/reliability/docs/application/#use-readiness-probe-to-detect-partial-unavailability"

    def check(self, namespaced_resources: NamespacedResources):

        offenders = []

        for pod in namespaced_resources.pods:
            for container in pod.spec.containers:
                if not container.readiness_probe:
                    offenders.append(pod)

        self.result = Result(status=True, resource_type="Pod")
        if offenders:
            self.result = Result(
                status=False,
                resource_type="Pod",
                resources=[i.metadata.name for i in offenders],
                namespace=namespaced_resources.namespace,
            )


class check_liveness_probes(Rule):
    _type = "namespace_based"
    pillar = "reliability"
    section = "applications"
    message = "Define liveness probes for pods."
    url = "https://aws.github.io/aws-eks-best-practices/reliability/docs/application/#use-liveness-probe-to-remove-unhealthy-pods"

    def check(self, namespaced_resources: NamespacedResources):

        offenders = []

        for pod in namespaced_resources.pods:
            for container in pod.spec.containers:
                if not container.liveness_probe:
                    offenders.append(pod)

        self.result = Result(status=True, resource_type="Pod")
        if offenders:
            self.result = Result(
                status=False,
                resource_type="Pod",
                resources=[i.metadata.name for i in offenders],
                namespace=namespaced_resources.namespace,
            )
